/*
 * $Id$
 *
 * Copyright (C) INRIA Rhne-Alpes, 2003-2004
 * Copyright (C) CNR Pisa, 2005
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

package fr.inrialpes.exmo.align.impl;

import java.lang.ClassNotFoundException;
import java.util.Hashtable;
import java.util.HashSet;
import java.util.Enumeration;
import java.util.Set;
import java.util.Iterator;
import java.util.Collections;
import java.util.List;
import java.util.ArrayList;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.IOException;
import java.net.URI;

import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLEntity;
import org.semanticweb.owl.model.OWLException;

import org.semanticweb.owl.align.Alignment;
import org.semanticweb.owl.align.AlignmentException;
import org.semanticweb.owl.align.AlignmentVisitor;
import org.semanticweb.owl.align.Cell;
import org.semanticweb.owl.align.Relation;

/**
 * Represents an OWL ontology alignment. An ontology comprises a number of
 * collections. Each ontology has a number of classes, properties and
 * individuals, along with a number of axioms asserting information about those
 * objects.
 * 
 * @author Jrme Euzenat, David Loup
 * @author Raphal Troncy
 */

public class MultipleAlignment extends BasicAlignment {

    public MultipleAlignment() {
	super();
    }

    public int nbCells() {
	// Raph: this is now wrong: I need multi-enumeration
	int sum = 0;
	for ( Enumeration e = getElements(); e.hasMoreElements(); ) {
	    sum = sum + ((HashSet)e.nextElement()).size();
	}
	return sum;
    }

    public Enumeration getElements1() { 
	// Raph: The elements are HashSet that contain the cells
	return hash1.elements();
    }

    public Enumeration getElements2() { 
	// Raph: The elements are HashSet that contain the cells
	return hash2.elements();
    }

    public Enumeration getKeys1() { 
	return hash1.keys();
    }

    public Enumeration getKeys2() { 
	return hash2.keys();
    }

    /*
     * Please note that all the following methods have been changed because they
     * considered that only ONE Entity can be aligned with another !!
     * A number of modifications have been introduced w.r.t. the BasicAlignment class
     */
    /** Cell methods **/
    public Cell addAlignCell(Object ob1, Object ob2, String relation, double measure) throws AlignmentException {

	if ( !( ob1 instanceof OWLEntity && ob2 instanceof OWLEntity ) )
	    throw new AlignmentException("addAlignCell: arguments must be OWLEntities");
	try {
	    Cell cell = (Cell) new BasicCell((OWLEntity) ob1, (OWLEntity) ob2, relation, measure);
	    HashSet s1 = (HashSet)hash1.get((Object)(((OWLEntity)ob1).getURI()));
	    if (s1 == null) {
		s1 = new HashSet();
		hash1.put((Object)(((OWLEntity)ob1).getURI()),s1);
	    }
	    s1.add(cell);
	    HashSet s2 = (HashSet)hash2.get((Object)(((OWLEntity)ob2).getURI()));
	    if (s2 == null) {
		s2 = new HashSet();
		hash2.put((Object)(((OWLEntity)ob2).getURI()),s2);
	    }
	    s2.add(cell);
	    return cell;
	} catch (OWLException e) {
	    throw new AlignmentException("getURI problem", e);
	}
    };

    public Cell addAlignCell(Object ob1, Object ob2) throws AlignmentException {
	return addAlignCell( ob1, ob2, "=", 1. );
    }

    // This is stuff from Raphael I am not fond of it (indexing by URIs)
    public Set getAlignCells1(URI u) {
	return (HashSet)hash1.get(u);
    }

    public Set getAlignCells2(URI u) {
	return (HashSet)hash2.get(u);
    }

    public Set getAlignCells1(OWLEntity o) {
	try{ return (HashSet)hash1.get(o.getURI());
	} catch (OWLException e) { return null;}
    }
    public Set getAlignCells2(OWLEntity o) {
	try{ return (HashSet)hash2.get(o.getURI());
	} catch (OWLException e) { return null;}
    }

    public void removeAlignCell(Cell c) throws AlignmentException {
	try {
	    HashSet s1 = (HashSet)hash1.get(((OWLEntity)c.getObject1()).getURI());
	    HashSet s2 = (HashSet)hash2.get(((OWLEntity)c.getObject2()).getURI());
	    s1.remove(c);
	    s2.remove(c);
	    if (s1.isEmpty())
		hash1.remove(((OWLEntity) c.getObject1()).getURI());
	    if (s2.isEmpty())
		hash2.remove(((OWLEntity) c.getObject2()).getURI());
	} catch (OWLException ex) {
	    throw new AlignmentException("getURI problem", ex);
	}
    }

    /***************************************************************************
     * The cut function suppresses from an alignment all the cell over a
     * particulat threshold
     **************************************************************************/
    public void cut2(double threshold) throws AlignmentException {
	for (Enumeration e = hash1.keys(); e.hasMoreElements(); ) {
	    HashSet s = (HashSet)hash1.get(e.nextElement());
	    for (Iterator i=s.iterator(); i.hasNext(); ) {
		Cell c = (Cell)i.next();
		if (c.getStrength() < threshold)
		    removeAlignCell( c );
	    }
	}
    };

    /***************************************************************************
     * The harden function acts like threshold but put all weights to 1.
     **************************************************************************/
    public void harden(double threshold) throws AlignmentException {
	for (Enumeration e = hash1.keys(); e.hasMoreElements();) {
	    HashSet s = (HashSet)e.nextElement();
	    for (Iterator i=s.iterator(); i.hasNext(); ) {
		Cell c = (Cell)i.next();
		if (c.getStrength() < threshold)
		    removeAlignCell( c );
		else
		    c.setStrength(1.);
	    }
	}
    };

    /***************************************************************************
     * Cut refinement :
     * - above n (hard)
     * - above n under the best ()
     * - getting the n% better (perc)
     * - getting the under n% of the best (prop)
     * - getting the n best values
     **************************************************************************/
    public void cut( String method, double threshold ) throws AlignmentException
    {
	// Check that threshold is a percent
	if ( threshold > 1. || threshold < 0. )
	    throw new AlignmentException( "Not a percentage or threshold : "+threshold );
	// Create a sorted list of cells
	// Raph: this will not work anymore
	List buffer = new ArrayList( hash1.values() );
	Collections.sort( buffer );
	int size = buffer.size();
	boolean found = false;
	int i = 0;
	// Depending on the method, find the limit
	if ( method.equals("hard") ){
	    for( i=0; i < size && !found ; i++ ) {
		if ( ((Cell)buffer.get(i)).getStrength() <= threshold ) found = true;
	    }
	} else if ( method.equals("span") ){
	    double max = ((Cell)buffer.get(0)).getStrength() -threshold;
	    for( i=0; i < size && !found ; i++ ) {
		if ( ((Cell)buffer.get(i)).getStrength() <= max ) found = true;
	    }
	} else if ( method.equals("prop") ){
	    double max = ((Cell)buffer.get(0)).getStrength() * (1-threshold);
	    for( i=0; i < size && !found ; i++ ) {
		if ( ((Cell)buffer.get(i)).getStrength() <= max ) found = true;
	    }
	} else if ( method.equals("perc") ){
	    i = (new Double(size*threshold)).intValue();
	} else if ( method.equals("best") ){
	    i=(new Double(threshold)).intValue();
	} else throw new AlignmentException( "Not a cut specification : "+method );
	// Flush the structure
	for( size-- ; size > i ; size-- ) buffer.remove(size);
	// Introduce the result back in the structure
	size = i;
	hash1.clear();
	hash2.clear();
	try {
	    for( i = 0; i < size; i++ ) {
		Cell c = (Cell)buffer.get(i);
		hash1.put((Object) (((OWLEntity)c.getObject1()).getURI()), c);
		hash2.put((Object) (((OWLEntity)c.getObject2()).getURI()), c);
	    }
	} catch (OWLException e) {
	    throw new AlignmentException("getURI problem", e);
	}
    };

   /**
     * The second alignment is meet with the first one meaning that for
     * any pair (o, o', n, r) in O and (o, o', n', r) in O' the resulting
     * alignment will contain:
     * ( o, o', meet(n,n'), r)
     * any pair which is in only one alignment is preserved.
     */
    public Alignment meet(Alignment align) throws AlignmentException {
	BasicAlignment result = new BasicAlignment();
	result.init(onto1,onto2);
	return result;
    }

   /**
     * The second alignment is join with the first one meaning that for
     * any pair (o, o', n, r) in O and (o, o', n', r) in O' the resulting
     * alignment will contain:
     * ( o, o", join(n,n'), r)
     * any pair which is in only one alignment is discarded.
     */
    public Alignment join(Alignment align) throws AlignmentException {
	BasicAlignment result = new BasicAlignment();
	result.init(onto1,onto2);
	return result;
    }

    /**
     * The second alignment is composed with the first one meaning that for
     * any pair (o, o', n, r) in O and (o',o", n', r') in O' the resulting
     * alignment will contain:
     * ( o, o", join(n,n'), compose(r, r')) iff compose(r,r') exists.
     */
    public Alignment compose(Alignment align) throws AlignmentException {
	BasicAlignment result = new BasicAlignment();
	result.init(onto1,onto2);
	return result;
    }

    /**
     * A new alignment is created such that for
     * any pair (o, o', n, r) in O the resulting alignment will contain:
     * ( o', o, n, inverse(r)) iff compose(r) exists.
     */
    public void inverse () {
	
		OWLOntology o = onto1;
		onto1 = onto2;
		onto2 = o;
		// We must inverse getType
		URI u = uri1;
		uri1 = uri2;
		uri2 = u;
		Hashtable h = hash1;
		hash1 = hash2;
		hash2 = h;
		for ( Enumeration e = getElements1() ; e.hasMoreElements(); ) {
			// TO CHANGE
			//((Cell)e.nextElement()).inverse();
		}
    };

    /** Housekeeping **/
    public void dump(ContentHandler h) {
    };

    /**
     * Incorporate the cell of the alignment into it own alignment. Note: for
     * the moment, this does not copy but really incorporates. So, if hardening
     * or cutting, are applied, then the ingested alignmment will be modified as
     * well.
     */
	protected void ingest(MultipleAlignment alignment) throws AlignmentException {

		for (Enumeration e = alignment.getKeys1(); e.hasMoreElements();) {
			Object o = (Object)e.nextElement();
			Set allCells = alignment.getAlignCells1((URI)o);
			HashSet s = (HashSet)hash1.get(o);
			if (s == null) {
				s = new HashSet();
				hash1.put(o,s);
				s.addAll(allCells);
			}
			/*
			else {
				for (Iterator i=allCells.iterator(); i.hasNext(); ) {
					BasicCell c1 = (BasicCell)i.next();
					boolean found = false;
					for (Iterator j=s.iterator(); j.hasNext(); ) {
						BasicCell c2 = (BasicCell)j.next();
						found = c1.equals(c2);
					}
					if (!found)
						s.add(c1);
				}
			}*/
		}
		for (Enumeration e = alignment.getKeys2(); e.hasMoreElements();) {
			Object o = (Object)e.nextElement();
			Set allCells = alignment.getAlignCells2((URI)o);
			HashSet s = (HashSet)hash2.get(o);
			if (s == null) {
				s = new HashSet();
				hash2.put(o,s);
				s.addAll(allCells);
			}
			/*
			else {
				for (Iterator i=allCells.iterator(); i.hasNext(); ) {
					BasicCell c1 = (BasicCell)i.next();
					boolean found = false;
					for (Iterator j=s.iterator(); j.hasNext(); ) {
						BasicCell c2 = (BasicCell)j.next();
						found = c1.equals(c2);
					}
					if (!found)
						s.add(c1);
				}
			}*/
		}
    };

    /**
     * This should be rewritten in order to generate the axiom ontology instead
     * of printing it! And then use ontology serialization for getting it
     * printed.
     */
    public void render( AlignmentVisitor renderer ) throws AlignmentException {
	accept(renderer);
    }
}
