/*
 * $Id: GroupEval.java,v 1.7 2005/08/30 21:51:01 euzenat Exp $
 *
 * Copyright (C) 2003 The University of Manchester
 * Copyright (C) 2003 The University of Karlsruhe
 * Copyright (C) 2003-2005, INRIA Rhne-Alpes
 * Copyright (C) 2004, Universit de Montral
 *
 * Modifications to the initial code base are copyright of their
 * respective authors, or their employers as appropriate.  Authorship
 * of the modifications may be determined from the ChangeLog placed at
 * the end of this file.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 */

/* This program evaluates the results of several ontology aligners in a row.
*/
package fr.inrialpes.exmo.align.util;

import org.semanticweb.owl.model.OWLOntology;

import org.semanticweb.owl.align.Alignment;
import org.semanticweb.owl.align.AlignmentProcess;
import org.semanticweb.owl.align.AlignmentVisitor;
import org.semanticweb.owl.align.Parameters;
import org.semanticweb.owl.align.Evaluator;

import fr.inrialpes.exmo.align.impl.BasicAlignment;
import fr.inrialpes.exmo.align.impl.BasicParameters;
import fr.inrialpes.exmo.align.impl.eval.PRGraphEvaluator;

import java.io.File;
import java.io.PrintStream;
import java.io.FileOutputStream;
import java.net.URI;
import java.lang.Double;
import java.lang.Integer;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.util.StringTokenizer;

import org.xml.sax.SAXException;

import gnu.getopt.LongOpt;
import gnu.getopt.Getopt;

import fr.inrialpes.exmo.align.parser.AlignmentParser;

/** A basic class for synthesizing the alignment results of an algorithm by a
    precision recall graph.
    
    <pre>
    java -cp procalign.jar fr.inrialpes.exmo.align.util.GenPlot [options]
    </pre>

    where the options are:
    <pre>
    -o filename --output=filename
    -d debug --debug=level
    -l list of compared algorithms
    -t output --type=output: xml/tex/html/ascii
   </pre>

   The input is taken in the current directory in a set of subdirectories (one per
   test) each directory contains a the alignment files (one per algorithm) for that test and the
   reference alignment file.

   If output is
   requested (<CODE>-o</CODE> flags), then output will be written to
    <CODE>output</CODE> if present, stdout by default.

<pre>
$Id: GroupEval.java,v 1.7 2005/08/30 21:51:01 euzenat Exp $
</pre>

@author Sean K. Bechhofer
@author Jrme Euzenat
    */

public class GenPlot {

    static Parameters params = null;
    static Vector listAlgo;
    static String fileNames = "";
    static String outFile = null; // This is unused for the moment
    static String type = "html";
    static int debug = 0;
    static Hashtable loaded = null;

    public static void main(String[] args) {
	try { run( args ); }
	catch (Exception ex) { ex.printStackTrace(); };
    }

    public static void run(String[] args) throws Exception {
	LongOpt[] longopts = new LongOpt[8];
	loaded = new Hashtable();

 	longopts[0] = new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h');
	longopts[1] = new LongOpt("output", LongOpt.REQUIRED_ARGUMENT, null, 'o');
	longopts[3] = new LongOpt("type", LongOpt.REQUIRED_ARGUMENT, null, 't');
	longopts[4] = new LongOpt("debug", LongOpt.OPTIONAL_ARGUMENT, null, 'd');
	longopts[6] = new LongOpt("list", LongOpt.REQUIRED_ARGUMENT, null, 'l');

	Getopt g = new Getopt("", args, "ho:d::l:t:", longopts);
	int c;
	String arg;

	while ((c = g.getopt()) != -1) {
	    switch (c) {
	    case 'h' :
		usage();
		return;
	    case 'o' :
		/* Write output here */
		outFile = g.getOptarg();
		break;
	    case 't' :
		/* Type of output (tex/html/xml/ascii) */
		type = g.getOptarg();
		break;
	    case 'l' :
		/* List of filename */
		fileNames = g.getOptarg();
		break;
	    case 'd' :
		/* Debug level  */
		arg = g.getOptarg();
		if ( arg != null ) debug = Integer.parseInt(arg.trim());
		else debug = 4;
		break;
	    }
	}

	// JE: StringTokenizer is obsoleted in Java 1.4 in favor of split: to change
	listAlgo = new Vector();
	StringTokenizer st = new StringTokenizer(fileNames,",");
	while (st.hasMoreTokens()) {
	    listAlgo.add(st.nextToken());
	}

	params = new BasicParameters();
	if (debug > 0) params.setParameter("debug", new Integer(debug-1));

	//print( iterateDirectories() );
    }
    /*
    public static Vector iterateDirectories (){
	Evaluator[] result = null;
	File [] subdir = null;
	try {
	    subdir = (new File(System.getProperty("user.dir"))).listFiles();
	} catch (Exception e) {
	    System.err.println("Cannot stat dir "+ e.getMessage());
	    usage();
	}
	int size = subdir.length;
	result = new Evaluator[size];
	int i = 0;
	for ( int j=0 ; j < size; j++ ) {
	    if( subdir[j].isDirectory() ) {
		if ( debug > 0 ) System.err.println("\nEntering directory "+subdir[j]);
		// eval the alignments in a subdirectory
		// store the result
		Object vect = (Object)iterateAlignments( subdir[j] );
		if ( vect != null ){
		    result.add(i, vect);
		    i++;
		}
		// result[i] = eval( subdir[j] );
		//	i++;
	    }
	}
	return result;
    }

    public static Vector iterateAlignments ( File dir ) {
	String prefix = dir.toURI().toString()+"/";
	Vector result = new Vector();
	boolean ok = false;
	result.add(0,(Object)dir.getName().toString());
	int i = 1;
	// for all alignments there,
	for ( Enumeration e = listAlgo.elements() ; e.hasMoreElements() ; i++) {
	    // call eval
	    // store the resul in a record
	    // return the record.
	    if ( debug > 1) System.err.println("  Considering result "+i);
	    Evaluator evaluator = (Evaluator)eval( prefix+"refalign.rdf", prefix+(String)e.nextElement()+".rdf");
	    if ( evaluator != null ) ok = true;
	    result.add( i, evaluator );
	}
	// Unload the ontologies.
	try {
	    for ( Enumeration e = loaded.elements() ; e.hasMoreElements();  ){
		OWLOntology o = (OWLOntology)e.nextElement();
		o.getOWLConnection().notifyOntologyDeleted( o );
	    }
	} catch (Exception ex) { System.err.println(ex); };
	if ( ok == true ) return result;
	else return (Vector)null;
    }

    public static Evaluator eval( String alignName1, String alignName2 ) {
	Evaluator eval = null;
	try {
	    int nextdebug;
	    if ( debug < 2 ) nextdebug = 0;
	    else nextdebug = debug - 2;
	    // Load alignments
	    AlignmentParser aparser1 = new AlignmentParser( nextdebug );
	    Alignment align1 = aparser1.parse( alignName1, loaded );
	    if ( debug > 1 ) System.err.println(" Alignment structure1 parsed");
	    AlignmentParser aparser2 = new AlignmentParser( nextdebug );
	    Alignment align2 = aparser2.parse( alignName2, loaded );
	    if ( debug > 1 ) System.err.println(" Alignment structure2 parsed");
	    // Create evaluator object
	    eval = new PRGraphEvaluator( align1, align2 );
	    // Compare
	    params.setParameter( "debug", new Integer( nextdebug ) );
	    eval.eval( params ) ;
	} catch (Exception ex) { System.err.println(ex); }

	// Unload the ontologies.
	try {
	    for ( Enumeration e = loaded.elements() ; e.hasMoreElements();  ){
		OWLOntology o = (OWLOntology)e.nextElement();
		o.getOWLConnection().notifyOntologyDeleted( o );
	    }
	} catch (Exception ex) { System.err.println(ex); };
	return eval;
	}*/


    /**
     * This does average plus plot
     *//*
    public static void print( Vector result ) {
	Vector final = new Vector();
	int i = 0;
	for ( Enumeration e = listAlgo.elements() ; e.hasMoreElements() ; i++) {
	    ...
	}
	
	for ( Enumeration e = listAlgo.elements() ; e.hasMoreElements() ; i++) {
	    System.out.println("\t"+(String)e.nextElement());
	    }
	int STEP = 10; // This should go upper level
	double[] plot = new double[STEP+1];
	int k = 0;
	// iterate for each test
	for( int i = result.length-1; i >= 0; i-- ){
	    PRGraphEvaluator ev = (PRGraphEvaluator)result[i];
	    if( ev != null ){
		k++;
		for( int j = 0; j <= STEP ; j++ ){
		    plot[j] = plot[j] + ev.getPrecision(j);
		}
	    }
	}
	// averaging (not a very good method anyway)
	for( int j = 0; j <= STEP ; j++ ){
	    plot[j] = plot[j]/k;
	}
	// printing for gnuplot
	for( int j = 0; j <= STEP ; j++ ){
	    System.out.println( (double)j/STEP +" "+plot[j] );
	}
    }

    // Borrowed and enhanced from
    // http://acm.sus.mcgill.ca/20020323/work/acm-19/B.j
    // What a pity that it is not in Java... (wait for 1.5)
    public static void printFormat(PrintStream w, double f){
	// JE: Must add the test is the value is Not a number, print NaN.
	if ( f != f ) {
	    w.print("NaN");
	} else {
	    int tmp = (int)(f*100);
	    int dec = tmp%100;
	    if( (int)(f*1000)%10 >= 5 ) dec++;
	    tmp /= 100;
	    w.print("" + tmp + ".");
	    if(dec < 10) w.print("0");
	    w.print("" + dec);
	}
    }*/

    public static void usage() {
	System.out.println("usage: GenPlot [options]");
	System.out.println("options are:");
	System.out.println("\t--type=html|xml|tex|ascii -t html|xml|tex|ascii\tSpecifies the output format");
	System.out.println("\t--list=algo1,...,algon -l algo1,...,algon\tSequence of the filenames to consider");
	System.out.println("\t--debug[=n] -d [n]\t\tReport debug info at level n");
	System.out.println("\t--help -h\t\t\tPrint this message");
    }
}
